"""
mq-sendmail

Usage:
    mq-sendmail [options] [<recipients>...]

<recipients> is only optional when "-t" is used and the message contains at
least one email address in the "To", "CC" or "BCC" header.

Options:
  --aliases=<ALIAS_FN>  Path to aliases file
  -C, --config=<CFG>    Path to the config file
  -f, --from=address    set envelope from address
  --set-date-header     Add a "Date" header to the message (if not present)
  --set-from-header     Add "From:" header to the message (if not present)
  --set-msgid-header    Add a "Message-ID" header to the message (if not present)
  --set-to-header       add a "To" header to the message (if not present)
  -t, --read-recipients   read additional recipients from the message
  --verbose, -v         more verbose program output
"""

import email.utils
import sys
from datetime import datetime as DateTime, timezone
from email.message import EmailMessage
from email.parser import BytesHeaderParser
from typing import Sequence

from docopt import docopt, printable_usage

from schwarz.mailqueue.aliases_parser import _parse_aliases, lookup_adresses
from schwarz.mailqueue.app_helpers import guess_config_path, init_app, init_smtp_mailer
from schwarz.mailqueue.message_handler import InMemoryMsg, MessageHandler
from schwarz.mailqueue.queue_runner import MaildirBackend


__all__ = ['mq_sendmail_main']

def mq_sendmail_main(argv=sys.argv, return_rc_code=False):
    arguments = docopt(__doc__, argv=argv[1:])
    config_path = guess_config_path(arguments['--config'])
    recipient_params = arguments['<recipients>']
    verbose = arguments['--verbose']

    aliases_fn = arguments['--aliases']
    envelope_from = arguments['--from']
    set_date_header = arguments['--set-date-header']
    set_from_header = arguments['--set-from-header']
    set_msgid_header = arguments['--set-msgid-header']
    set_to_header = arguments['--set-to-header']
    read_recipients = arguments['--read-recipients']

    if not recipient_params and not read_recipients:
        usage_str = printable_usage(__doc__)
        sys.stdout.write(usage_str + '\n')
        sys.stderr.write('At least one recipient address is required.\n')
        sys.exit(1)

    msg_bytes = sys.stdin.buffer.read()
    input_headers = BytesHeaderParser().parsebytes(msg_bytes)
    aliases = _parse_aliases(aliases_fn) if aliases_fn else None
    if read_recipients:
        msg_recipients = _recipients_from_message(input_headers)
    else:
        msg_recipients = None
    recipients = lookup_adresses(recipient_params, aliases, msg_recipients=msg_recipients)
    if not recipients:
        sys.stderr.write('No recipient addresses found in message.\n')
        sys.exit(1)

    cli_options = {
        'verbose': verbose,
        'quiet'  : not verbose,
    }
    settings = init_app(config_path, options=cli_options)
    if envelope_from:
        msg_sender = envelope_from
    else:
        msg_sender = settings.get('from')
    if not msg_sender:
        sys.stderr.write('No envelope sender address given (use "--from=...").\n')
        sys.exit(81)

    extra_header_lines = autogenerated_headers(
        input_headers,
        set_date_header,
        set_from_header,
        set_msgid_header,
        set_to_header,
        msg_sender,
        recipients,
    )
    msg = InMemoryMsg(msg_sender, recipients, extra_header_lines + msg_bytes)

    transports = [init_smtp_mailer(settings)]
    queue_dir = settings.get('queue_dir')
    if queue_dir:
        transports.append(MaildirBackend(queue_dir))
    mh = MessageHandler(transports=transports)
    send_result = mh.send_message(msg)

    if verbose:
        cli_output = build_cli_output(send_result)
        print(cli_output)
    was_sent = bool(send_result)
    exit_code = 0 if was_sent else 100
    if return_rc_code:
        return exit_code
    sys.exit(exit_code)


def _recipients_from_message(input_headers) -> Sequence[str]:
    recipients = set()
    for header in {'To', 'CC', 'BCC'}:
        header_lines = input_headers.get_all(header)
        if header_lines is None:
            continue
        for (_, recipient_addr) in email.utils.getaddresses(header_lines):
            recipients.add(recipient_addr)
    return tuple(recipients)


def autogenerated_headers(
        input_headers,
        set_date_header, set_from_header, set_msgid_header,
        set_to_header,
        msg_sender,
        recipients,
    ) -> bytes:
    extra_headers = EmailMessage()
    input_msg_date = input_headers.get('Date')
    if set_date_header and not input_msg_date:
        extra_headers['Date'] = email.utils.format_datetime(DateTime.now(timezone.utc))
    input_msg_from = input_headers.get('From')
    if set_from_header and not input_msg_from:
        extra_headers['From'] = msg_sender
    input_msg_to = input_headers.get('To')
    if set_to_header and (not input_msg_to) and recipients:
        extra_headers['To'] = ', '.join(recipients)
    input_msg_id = input_headers.get('Message-ID')
    if set_msgid_header and not input_msg_id:
        _, smtp_sender_domain = msg_sender.split('@', 1)
        extra_headers['Message-ID'] = email.utils.make_msgid(domain=smtp_sender_domain)

    prepended_header_lines = b''
    if extra_headers:
        fake_msg_bytes = extra_headers.as_bytes()
        prepended_header_lines = fake_msg_bytes.split(b'\n\n', 1)[0] + b'\n'
    return prepended_header_lines


def build_cli_output(send_result) -> str:
    was_sent = bool(send_result)
    if was_sent:
        if send_result.queued:
            verb = 'queued'
            via = f' via {send_result.transport}'
        elif send_result.discarded:
            verb = 'discarded'
            via = ''
        else:
            verb = 'sent'
            via = f' via {send_result.transport}'
        return f'Message was {verb}{via}.'
    return ''
